from typing import Optional
from datetime import date
from pydantic import BaseModel
from nport.xmltools import child_text, child_value


class Registrant(BaseModel):
    # Elements of A.1
    name: str
    file_number: str
    cik: int
    lei: str
    street_address1: str
    street_address2: Optional[str]
    city: str
    state: Optional[str]
    country: Optional[str]
    zip_code: str
    telephone: str

    # Elements of A.2
    series_name: str
    series_edgar_identifier: str
    series_lei: str

    # Elements of A.3 and A.4
    fiscal_year_end: date
    report_date: date
    is_final_filing: bool

    def __repr__(self):
        return f"{self.name} [{self.report_date:%Y-%m-%d}]"

    def __str__(self):
        return self.__repr__()

    @classmethod
    def from_xml(cls, tag):
        name = child_text(tag, "regName")
        file_number = child_text(tag, "regFileNumber")
        cik = child_text(tag, "regCik")
        lei = child_text(tag, "regLei")
        street_address1 = child_text(tag, "regStreet1")
        street_address2 = child_text(tag, "regStreet2")
        city = child_text(tag, "regCity")
        state = child_value(tag, "regStateConditional", "regState")
        country = child_value(tag, "regStateConditional", "regCountry")
        zip_code = child_text(tag, "regZipOrPostalCode")
        telephone = child_text(tag, "regPhone")

        series_name = child_text(tag, "seriesName")
        series_edgar_identifier = child_text(tag, "seriesId")
        series_lei = child_text(tag, "seriesLei")

        fiscal_year_end = child_text(tag, "repPdEnd")
        report_date = child_text(tag, "repPdDate")
        is_final_filing = child_text(tag, "isFinalFiling") == "Y"

        return cls(
            name=name,
            file_number=file_number,
            cik=cik,
            lei=lei,
            street_address1=street_address1,
            street_address2=street_address2,
            city=city,
            state=state,
            country=country,
            zip_code=zip_code,
            telephone=telephone,
            series_name=series_name,
            series_edgar_identifier=series_edgar_identifier,
            series_lei=series_lei,
            fiscal_year_end=fiscal_year_end,
            report_date=report_date,
            is_final_filing=is_final_filing
        )


class PortfolioInterestRisk(BaseModel):
    # Elements of B.3.a and B.3.b
    currency: str
    dv01_3m: float
    dv01_1y: float
    dv01_5y: float
    dv01_10y: float
    dv01_30y: float
    dv100_3m: float
    dv100_1y: float
    dv100_5y: float
    dv100_10y: float
    dv100_30y: float

    @classmethod
    def from_xml(cls, tag):
        currency = child_text(tag, "curCd")
        dv01_3m = child_value(tag, "intrstRtRiskdv01", "period3Mon")
        dv01_1y = child_value(tag, "intrstRtRiskdv01", "period1Yr")
        dv01_5y = child_value(tag, "intrstRtRiskdv01", "period5Yr")
        dv01_10y = child_value(tag, "intrstRtRiskdv01", "period10Yr")
        dv01_30y = child_value(tag, "intrstRtRiskdv01", "period30Yr")
        dv100_3m = child_value(tag, "intrstRtRiskdv100", "period3Mon")
        dv100_1y = child_value(tag, "intrstRtRiskdv100", "period1Yr")
        dv100_5y = child_value(tag, "intrstRtRiskdv100", "period5Yr")
        dv100_10y = child_value(tag, "intrstRtRiskdv100", "period10Yr")
        dv100_30y = child_value(tag, "intrstRtRiskdv100", "period30Yr")

        return cls(
            currency=currency,
            dv01_3m=dv01_3m,
            dv01_1y=dv01_1y,
            dv01_5y=dv01_5y,
            dv01_10y=dv01_10y,
            dv01_30y=dv01_30y,
            dv100_3m=dv100_3m,
            dv100_1y=dv100_1y,
            dv100_5y=dv100_5y,
            dv100_10y=dv100_10y,
            dv100_30y=dv100_30y
        )


class PortfolioCreditRisk(BaseModel):
    # Elements of B.3.c
    cs01_3m_ig: float
    cs01_1y_ig: float
    cs01_5y_ig: float
    cs01_10y_ig: float
    cs01_30y_ig: float
    cs01_3m_junk: float
    cs01_1y_junk: float
    cs01_5y_junk: float
    cs01_10y_junk: float
    cs01_30y_junk: float

    @classmethod
    def from_xml(cls, tag):
        cs01_3m_ig = child_value(tag, "creditSprdRiskInvstGrade", "period3Mon")
        cs01_1y_ig = child_value(tag, "creditSprdRiskInvstGrade", "period1Yr")
        cs01_5y_ig = child_value(tag, "creditSprdRiskInvstGrade", "period5Yr")
        cs01_10y_ig = child_value(tag, "creditSprdRiskInvstGrade", "period10Yr")
        cs01_30y_ig = child_value(tag, "creditSprdRiskInvstGrade", "period30Yr")
        cs01_3m_junk = child_value(tag, "creditSprdRiskNonInvstGrade", "period3Mon")
        cs01_1y_junk = child_value(tag, "creditSprdRiskNonInvstGrade", "period1Yr")
        cs01_5y_junk = child_value(tag, "creditSprdRiskNonInvstGrade", "period5Yr")
        cs01_10y_junk = child_value(tag, "creditSprdRiskNonInvstGrade", "period10Yr")
        cs01_30y_junk = child_value(tag, "creditSprdRiskNonInvstGrade", "period30Yr")

        return cls(
            cs01_3m_ig=cs01_3m_ig,
            cs01_1y_ig=cs01_1y_ig,
            cs01_5y_ig=cs01_5y_ig,
            cs01_10y_ig=cs01_10y_ig,
            cs01_30y_ig=cs01_30y_ig,
            cs01_3m_junk=cs01_3m_junk,
            cs01_1y_junk=cs01_1y_junk,
            cs01_5y_junk=cs01_5y_junk,
            cs01_10y_junk=cs01_10y_junk,
            cs01_30y_junk=cs01_30y_junk
        )


class Borrower(BaseModel):
    # Elements of B.4.a
    name: str
    lei: Optional[str]
    aggregate_value: float


class NonCashCollateral(BaseModel):
    # Elements of B.4.b
    aggregate_principal: float
    aggregate_value: float
    asset_type: str


class ClassReturn(BaseModel):
    # Elements of B.5.a
    return_month1: float
    return_month2: float
    return_month3: float
    class_identification: str


class Fund(BaseModel):
    # Elements of B.1
    total_assets: float
    total_liabilities: float
    net_assets: float

    # Elements of B.2
    misc_assets: float
    controlled_foreign_corporation_assets: float
    borrowings_short_bank: float
    borrowings_short_companies: float
    borrowings_short_affiliates: float
    borrowings_short_others: float
    borrowings_medium_bank: float
    borrowings_medium_companies: float
    borrowings_medium_affiliates: float
    borrowings_medium_others: float
    payables_delayed: float
    payables_standby: float
    liquidation_preference: float
    cash: float

    # Elements of B.3
    interest_risk: list[PortfolioInterestRisk]
    credit_risk: Optional[PortfolioCreditRisk]

    # Elements of B.4
    borrowers: list[Borrower]
    is_non_cash_collateral: bool
    non_cash_collateral: list[NonCashCollateral]

    # Elements of B.5
    class_returns: list[ClassReturn]

    def __repr__(self):
        return f"Net Assets: ${self.net_assets:,.2f}"

    def __str__(self):
        return self.__repr__()

    @classmethod
    def from_xml(cls, tag):
        total_assets = child_text(tag, "totAssets")
        total_liabilities = child_text(tag, "totLiabs")
        net_assets = child_text(tag, "netAssets")

        misc_assets = child_text(tag, "assetsAttrMiscSec")
        controlled_foreign_corporation_assets = child_text(tag, "assetsInvested")
        borrowings_short_bank = child_text(tag, "amtPayOneYrBanksBorr")
        borrowings_short_companies = child_text(tag, "amtPayOneYrCtrldComp")
        borrowings_short_affiliates = child_text(tag, "amtPayOneYrOthAffil")
        borrowings_short_others = child_text(tag, "amtPayOneYrOther")
        borrowings_medium_bank = child_text(tag, "amtPayAftOneYrBanksBorr")
        borrowings_medium_companies = child_text(tag, "amtPayAftOneYrCtrldComp")
        borrowings_medium_affiliates = child_text(tag, "amtPayAftOneYrOthAffil")
        borrowings_medium_others = child_text(tag, "amtPayAftOneYrOther")
        payables_delayed = child_text(tag, "delayDeliv")
        payables_standby = child_text(tag, "standByCommit")
        liquidation_preference = child_text(tag, "liquidPref")
        cash = child_text(tag, "cshNotRptdInCorD")

        interest_risk = []
        interest_risks_tag = tag.find("CurMetrics")
        if interest_risks_tag:
            for interest_risk_tag in interest_risks_tag.find_all("curMetric"):
                interest_risk.append(PortfolioInterestRisk.from_xml(interest_risk_tag))

        credit_risk = None
        if tag.find("creditSprdRiskInvstGrade") or tag.find("creditSprdRiskNonInvstGrade"):
            credit_risk = PortfolioCreditRisk.from_xml(tag)

        borrowers = []
        borrowers_tag = tag.find("borrowers")
        if borrowers_tag:
            for borrower_tag in borrowers_tag.find_all("borrower"):
                borrowers.append(Borrower(
                    name=borrower_tag.attrs.get("name"),
                    lei=borrower_tag.attrs.get("lei"),
                    aggregate_value=borrower_tag.attrs.get("aggrVal")
                ))

        is_non_cash_collateral = child_text(tag, "isNonCashCollateral") == "Y"
        non_cash_collateral = []
        non_cash_collaterals_tag = tag.find("aggregateInfos")
        if non_cash_collaterals_tag:
            is_non_cash_collateral = True
            for non_cash_collateral_tag in non_cash_collaterals_tag.find_all("aggregateInfo"):
                non_cash_collateral.append(NonCashCollateral(
                    aggregate_principal=non_cash_collateral_tag.attrs.get("amt"),
                    aggregate_value=non_cash_collateral_tag.attrs.get("collatrl"),
                    asset_type=child_text(non_cash_collateral_tag, "invstCat")
                ))

        class_returns = []
        return_info_tag = tag.find("returnInfo")
        total_return_infos_tag = return_info_tag.find("monthlyTotReturns")
        for total_return_info_tag in total_return_infos_tag.find_all("monthlyTotReturn"):
            class_returns.append(ClassReturn(
                return_month1=total_return_info_tag.attrs.get("rtn1"),
                return_month2=total_return_info_tag.attrs.get("rtn2"),
                return_month3=total_return_info_tag.attrs.get("rtn3"),
                class_identification=total_return_info_tag.attrs.get("classId")
            ))

        return cls(
            total_assets=total_assets,
            total_liabilities=total_liabilities,
            net_assets=net_assets,
            misc_assets=misc_assets,
            controlled_foreign_corporation_assets=controlled_foreign_corporation_assets,
            borrowings_short_bank=borrowings_short_bank,
            borrowings_short_companies=borrowings_short_companies,
            borrowings_short_affiliates=borrowings_short_affiliates,
            borrowings_short_others=borrowings_short_others,
            borrowings_medium_bank=borrowings_medium_bank,
            borrowings_medium_companies=borrowings_medium_companies,
            borrowings_medium_affiliates=borrowings_medium_affiliates,
            borrowings_medium_others=borrowings_medium_others,
            payables_delayed=payables_delayed,
            payables_standby=payables_standby,
            liquidation_preference=liquidation_preference,
            cash=cash,
            interest_risk=interest_risk,
            credit_risk=credit_risk,
            borrowers=borrowers,
            is_non_cash_collateral=is_non_cash_collateral,
            non_cash_collateral=non_cash_collateral,
            class_returns=class_returns
        )



from typing import Optional, Union
from decimal import Decimal
from bs4 import Tag, BeautifulSoup


__all__ = [
    'child_text',
    'child_value'
]


def find_element(
        xml_tag_or_string: Union[str, BeautifulSoup, Tag], element_name) -> Tag:
    """
    Find the element with that name in the string or Tag
    :param xml_tag_or_string: either an exml tag or string containing xml
    :param element_name: The name of the element to find
    :return: An element
    """
    if isinstance(xml_tag_or_string, Tag):
        return xml_tag_or_string.find(element_name, recursive=False)
    elif isinstance(xml_tag_or_string, str) and "<" in xml_tag_or_string:
        soup: BeautifulSoup = BeautifulSoup(xml_tag_or_string, features="xml")
        return soup.find(element_name, recursive=False)


def child_text(parent: Tag, child: str) -> Optional[str]:
    """
    Get the text of the child element if it exists or None
    :param parent: The parent tag
    :param child: The name of the child element
    :return: the text of the child element if it exists or None
    """
    el = parent.find(child, recursive=False)
    if el and isinstance(el.text, str) and el.text.strip() not in ("N/A", "000000000", "XXXX"):
        return el.text.strip()
    return None


def child_value(parent: Tag, child: str, key: str = "value") -> Optional[str]:
    """
    Get the value of the child element if it exists or None
    :param parent: The parent tag
    :param child: The name of the child element
    :param key: The key of the child element's attributes
    :return: the value of the child element attribute if it exists or None
    """
    el = parent.find(child, recursive=False)
    if el:
        value = el.attrs.get(key)
        if isinstance(value, str) and value.strip() not in ("N/A", "000000000", "XXXX"):
            return value
    return None



import os.path

from pydantic import BaseModel
from bs4 import BeautifulSoup
import pandas as pd
import xmlschema
from nport.fund import Registrant, Fund
from nport.instruments import BaseInstrument, DebtSecurity, Derivative


PATH, _ = os.path.split(__file__)


class NPORT(BaseModel):
    registrant: Registrant
    fund: Fund
    securities: list[BaseInstrument | DebtSecurity | Derivative]

    def __repr__(self):
        return f"{self.registrant.name} [{self.registrant.report_date:%Y-%m-%d}]"

    def __str__(self):
        return self.__repr__()

    @classmethod
    def from_file(cls, path: str, **kwargs):
        with open(path, "r") as file:
            xml = file.read()

        validate = kwargs.get("validate", True)
        if validate:
            schema = xmlschema.XMLSchema(os.path.join(PATH, "..", "schema", "nport", "eis_NPORT_Filer.xsd"))
            try:
                xmlschema.validate(xml, schema)
            except xmlschema.XMLSchemaDecodeError:
                pass

        root = BeautifulSoup(xml, features="xml")
        return cls.from_xml(root)

    @classmethod
    def from_str(cls, xml: str, **kwargs):
        validate = kwargs.get("validate", True)
        if validate:
            schema = xmlschema.XMLSchema(os.path.join(PATH, "..", "schema", "nport", "eis_NPORT_Filer.xsd"))
            try:
                xmlschema.validate(xml, schema)
            except xmlschema.XMLSchemaDecodeError:
                pass

        root = BeautifulSoup(xml, features="xml")
        return cls.from_xml(root)

    @classmethod
    def from_xml(cls, root):
        form_tag = root.find("formData")

        registrant_tag = form_tag.find("genInfo")
        registrant = Registrant.from_xml(registrant_tag)

        fund_tag = form_tag.find("fundInfo")
        fund = Fund.from_xml(fund_tag)

        securities_tag = form_tag.find("invstOrSecs")
        securities = []
        for security_tag in securities_tag.find_all("invstOrSec"):
            if security_tag.find("debtSec"):
                securities.append(DebtSecurity.from_xml(security_tag, registrant.report_date))
            elif security_tag.find("derivativeInfo"):
                securities.append(Derivative.from_xml(security_tag, registrant.report_date))
            else:
                securities.append(BaseInstrument.from_xml(security_tag, registrant.report_date))

        return cls(
            registrant=registrant,
            fund=fund,
            securities=securities
        )

    def export_prices(self):
        price_list = [security.to_list() for security in self.securities]
        return pd.DataFrame(price_list)



from typing import Optional
from datetime import date
from decimal import Decimal
from pydantic import BaseModel
import pandas as pd
from nport.xmltools import child_value, child_text


class BaseInstrument(BaseModel):
    # Elements of C.1
    issuer: Optional[str]
    issuer_lei: Optional[str]
    name: str
    cusip: Optional[str]
    isin: Optional[str]
    ticker: Optional[str]
    other_identifier_name: Optional[str]
    other_identifier: Optional[str]

    # Elements of C.2 and C.3
    balance: Decimal
    value_usd: Decimal
    value_local: Decimal
    price: Decimal
    units: str
    units_description: Optional[str]
    currency: str
    exchange_rate: Optional[Decimal]
    fund_percentage: Decimal
    payoff_profile: Optional[str]

    # Elements of C.4, C.5 and C.6
    asset_type: str
    asset_description: Optional[str]
    issuer_type: str
    issuer_description: Optional[str]
    issuer_country: str
    issuer_country_exposure: Optional[str]
    is_restricted_security: bool

    # Elements of C.7 and C.8
    liquidity: Optional[int]
    fair_value_level: Optional[int]

    # Elements of C.12
    is_cash_collateral_value: bool
    cash_collateral_value: Optional[Decimal]
    is_non_cash_collateral: bool
    non_cash_collateral_value: Optional[Decimal]
    is_on_loan: bool
    on_loan_value: Optional[Decimal]

    # Additional from A.3
    report_date: date

    def __repr__(self):
        return f"{self.name} [{self.issuer}]"

    def __str__(self):
        return self.__repr__()

    @classmethod
    def from_xml(cls, tag, report_date):
        issuer = child_text(tag, "name")
        issuer_lei = child_text(tag, "lei")
        name = child_text(tag, "title")
        cusip = child_text(tag, "cusip")

        identifiers_tag = tag.find("identifiers")
        ticker = child_value(identifiers_tag, "ticker")
        isin = child_value(identifiers_tag, "isin")
        other_identifier_name = child_value(identifiers_tag, "other", "otherDesc")
        other_identifier = child_value(identifiers_tag, "other")

        balance = child_text(tag, "balance")
        value_usd = child_text(tag, "valUSD")
        units = child_text(tag, "units")
        units_description = child_text(tag, "descOthUnits")
        currency = child_text(tag, "curCd")
        exchange_rate = None
        if currency is None:
            currency = child_value(tag, "currencyConditional", "curCd")
            exchange_rate = child_value(tag, "currencyConditional", "exchangeRt")
        fund_percentage = child_text(tag, "pctVal")
        payoff_profile = child_text(tag, "payoffProfile")

        value_local = None
        if currency and currency != "USD" and exchange_rate:
            value_local = Decimal(exchange_rate) * Decimal(value_usd)
        elif currency and currency == "USD":
            value_local = value_usd
        price = Decimal(value_local) / Decimal(balance)
        if units == "PA":
            price = 100 * price
        if payoff_profile == "Short":
            price = -1 * price

        asset_type = child_text(tag, "assetCat")
        asset_description = None
        if asset_type is None:
            asset_type = child_value(tag, "assetConditional", "assetCat")
            asset_description = child_value(tag, "assetConditional", "desc")
        issuer_type = child_text(tag, "issuerCat")
        issuer_description = None
        if issuer_type is None:
            issuer_type = child_value(tag, "issuerConditional", "issuerCat")
            issuer_description = child_value(tag, "issuerConditional", "desc")
        issuer_country = child_text(tag, "invCountry")
        issuer_country_exposure = child_text(tag, "invOthCountry")
        is_restricted_security = child_text(tag, "isRestrictedSec") == "Y"

        liquidity = child_text(tag, "fundCat")
        fair_value_level = child_text(tag, "fairValLevel")

        security_lending_tag = tag.find("securityLending")
        is_cash_collateral_value = child_text(security_lending_tag, "isCashCollateral")
        cash_collateral_value = None
        if is_cash_collateral_value is None:
            is_cash_collateral_value = child_value(security_lending_tag, "cashCollateralCondition", "isCashCollateral")
            cash_collateral_value = child_value(security_lending_tag, "cashCollateralCondition", "cashCollateralVal")
        is_cash_collateral_value = is_cash_collateral_value == "Y"

        is_non_cash_collateral = child_text(security_lending_tag, "isNonCashCollateral")
        non_cash_collateral_value = None
        if is_non_cash_collateral is None:
            is_non_cash_collateral = child_value(security_lending_tag, "nonCashCollateralCondition", "isNonCashCollateral")
            non_cash_collateral_value = child_value(security_lending_tag, "nonCashCollateralCondition", "nonCashCollateralVal")
        is_non_cash_collateral = is_non_cash_collateral == "Y"

        is_on_loan = child_text(security_lending_tag, "isLoanByFund")
        on_loan_value = None
        if is_on_loan is None:
            is_on_loan = child_value(security_lending_tag, "loanByFundCondition", "isLoanByFund")
            on_loan_value = child_value(security_lending_tag, "loanByFundCondition", "loanVal")
        is_on_loan = is_on_loan == "Y"

        return cls(
            issuer=issuer,
            issuer_lei=issuer_lei,
            name=name,
            cusip=cusip,
            isin=isin,
            ticker=ticker,
            other_identifier_name=other_identifier_name,
            other_identifier=other_identifier,
            balance=balance,
            value_usd=value_usd,
            value_local=value_local,
            price=price,
            units=units,
            units_description=units_description,
            currency=currency,
            exchange_rate=exchange_rate,
            fund_percentage=fund_percentage,
            payoff_profile=payoff_profile,
            asset_type=asset_type,
            asset_description=asset_description,
            issuer_type=issuer_type,
            issuer_description=issuer_description,
            issuer_country=issuer_country,
            issuer_country_exposure=issuer_country_exposure,
            is_restricted_security=is_restricted_security,
            liquidity=liquidity,
            fair_value_level=fair_value_level,
            is_cash_collateral_value=is_cash_collateral_value,
            cash_collateral_value=cash_collateral_value,
            is_non_cash_collateral=is_non_cash_collateral,
            non_cash_collateral_value=non_cash_collateral_value,
            is_on_loan=is_on_loan,
            on_loan_value=on_loan_value,
            report_date=report_date
        )

    def to_list(self):
        return pd.Series(
            [
                self.name,
                self.cusip,
                self.isin,
                self.ticker,
                self.other_identifier,
                self.other_identifier_name,
                self.issuer,
                self.issuer_description,
                self.issuer_lei,
                self.issuer_type,
                self.issuer_country,
                self.issuer_country_exposure,
                self.asset_type,
                self.asset_description,
                self.payoff_profile,
                self.currency,
                self.balance,
                self.value_local,
                self.exchange_rate,
                self.price,
                self.fair_value_level
            ],
            index=[
                "Name", "CUSIP", "ISIN", "Ticker", "OtherIdentifier", "OtherIdentifierName", "Issuer",
                "IssuerDescription", "IssuerLEI", "IssuerType", "IssuerCountry", "IssuerCountryExposure",
                "AssetType", "AssetTypeDescription", "PayoffProfile", "Currency", "Balance", "Value", "ExchangeRate",
                "Price", "FairValueLevel"
            ]
        )


class DebtSecurity(BaseInstrument):
    # Elements of C.9
    maturity_date: date
    coupon_type: Optional[str]
    coupon_rate: Optional[Decimal]
    is_default: bool
    is_deferred_interest: bool
    is_paid_in_kind: bool
    is_mandatory_convertible: Optional[bool]
    is_contingent_convertible: Optional[bool]
    reference_issuer: Optional[str]
    reference_name: Optional[str]
    reference_currency: Optional[str]
    reference_cusip: Optional[str]
    reference_isin: Optional[str]
    reference_ticker: Optional[str]
    reference_other_identifier_name: Optional[str]
    reference_other_identifier: Optional[str]
    conversion_ratio: Optional[Decimal]
    conversion_currency: Optional[str]
    delta: Optional[Decimal]

    @classmethod
    def from_xml(cls, tag, report_date):
        debt_security_tag = tag.find("debtSec")
        maturity_date = child_text(debt_security_tag, "maturityDt")
        coupon_type = child_text(debt_security_tag, "couponKind")
        coupon_rate = child_text(debt_security_tag, "annualizedRt")
        is_default = child_text(debt_security_tag, "isDefault") == "Y"
        is_deferred_interest = child_text(debt_security_tag, "areIntrstPmntsInArrs") == "Y"
        is_paid_in_kind = child_text(debt_security_tag, "isPaidKind")

        is_mandatory_convertible = child_text(debt_security_tag, "isMandatoryConvrtbl")
        if is_mandatory_convertible:
            is_mandatory_convertible = is_mandatory_convertible == "Y"
        is_contingent_convertible = child_text(debt_security_tag, "isContngtConvrtbl")
        if is_contingent_convertible:
            is_contingent_convertible = is_contingent_convertible == "Y"

        reference_issuer = None
        reference_name = None
        reference_currency = None
        reference_cusip = None
        reference_isin = None
        reference_ticker = None
        reference_other_identifier_name = None
        reference_other_identifier = None
        reference_tag = debt_security_tag.find("dbtSecRefInstruments")
        if reference_tag:
            reference_issuer = child_text(reference_tag, "name")
            reference_name = child_text(reference_tag, "title")
            reference_currency = child_text(reference_tag, "curCd")
            reference_identifier_tag = reference_tag.find("identifiers")
            reference_cusip = child_value(reference_identifier_tag, "cusip")
            reference_isin = child_value(reference_identifier_tag, "isin")
            reference_ticker = child_value(reference_identifier_tag, "ticker")
            reference_other_identifier_name = child_value(reference_identifier_tag, "other", "otherDesc")
            reference_other_identifier = child_value(reference_identifier_tag, "other")

        conversion_ratio = None
        conversion_currency = None
        currency_info_tag = debt_security_tag.find("currencyInfos")
        if currency_info_tag:
            conversion_ratio = child_value(currency_info_tag, "currencyInfo", "convRatio")
            conversion_currency = child_value(currency_info_tag, "currencyInfo", "curCd")
        delta = child_text(debt_security_tag, "delta")

        base = BaseInstrument.from_xml(tag, report_date)
        return cls(
            **base.__dict__,
            maturity_date=maturity_date,
            coupon_type=coupon_type,
            coupon_rate=coupon_rate,
            is_default=is_default,
            is_deferred_interest=is_deferred_interest,
            is_paid_in_kind=is_paid_in_kind,
            is_mandatory_convertible=is_mandatory_convertible,
            is_contingent_convertible=is_contingent_convertible,
            reference_issuer=reference_issuer,
            reference_name=reference_name,
            reference_currency=reference_currency,
            reference_cusip=reference_cusip,
            reference_isin=reference_isin,
            reference_ticker=reference_ticker,
            reference_other_identifier_name=reference_other_identifier_name,
            reference_other_identifier=reference_other_identifier,
            conversion_ratio=conversion_ratio,
            conversion_currency=conversion_currency,
            delta=delta
        )


class RepurchasementAgreement(BaseInstrument):
    # Elements of C.10
    agreement_category: str
    is_cleared_by_central_counterparty: bool
    counterparty_name: Optional[str]
    counterparty_lei: Optional[str]
    is_triparty: bool
    repurchase_rate: Decimal
    maturity_date: date
    principal: Decimal
    collateral: Decimal
    collateral_type: str


class Derivative(BaseInstrument):
    # Elements of C.11
    derivative_type: str
    derivative_type_description: Optional[str]
    counterparty_name: Optional[str]
    counterparty_lei: Optional[str]


    @classmethod
    def from_xml(cls, tag, report_date):
        derivative_tag = tag.find("derivativeInfo")
        base = BaseInstrument.from_xml(tag, report_date)

        counterparty_tag = derivative_tag.find("counterparties")
        counterparty_name = child_text(counterparty_tag, "counterpartyName")
        counterparty_lei = child_text(counterparty_tag, "counterpartyLei")

        derivative_type_description = None
        if derivative_tag.find("fwdDeriv"):
            forward_tag = derivative_tag.find("fwdDeriv")

            if forward_tag.find("payOffProf"):
                derivative_type = "Forward"
                base.payoff_profile = child_text(forward_tag, "payOffProf")

                reference_instrument_tag = forward_tag.find("descRefInstrmnt")
                reference_instrument = None
                if reference_instrument_tag:
                    if reference_instrument_tag.find("indexBasketInfo"):
                        reference_instrument = ReferenceIndex.from_xml(reference_instrument_tag.find("indexBasketInfo"))
                    elif reference_instrument_tag.find("nestedDerivInfo"):
                        reference_instrument = ReferenceDerivative.from_xml(
                            reference_instrument_tag.find("nestedDerivInfo"))
                    else:
                        reference_instrument = ReferenceOther.from_xml(reference_instrument_tag.find("otherRefInst"))

                maturity = child_text(forward_tag, "expDate")
                notional = child_text(forward_tag, "notionalAmt")
                derivative_currency = child_text(forward_tag, "curCd")
                unrealized_appreciation = child_text(forward_tag, "unrealizedAppr")
                return ForwardFuture(
                    **base.__dict__,
                    derivative_type=derivative_type,
                    derivative_type_description=derivative_type_description,
                    counterparty_name=counterparty_name,
                    counterparty_lei=counterparty_lei,
                    reference_instrument=reference_instrument,
                    maturity=maturity,
                    notional=notional,
                    derivative_currency=derivative_currency,
                    unrealized_appreciation=unrealized_appreciation
                )

            else:
                derivative_type = "FX Forward"
                currency_amount_sold = child_text(forward_tag, "amtCurSold")
                currency_sold = child_text(forward_tag, "curSold")
                currency_amount_purchased = child_text(forward_tag, "amtCurPur")
                currency_purchased = child_text(forward_tag, "curPur")
                settlement = child_text(forward_tag, "settlementDt")
                unrealized_appreciation = child_text(forward_tag, "unrealizedAppr")
                return FXForwardOrSwap(
                    **base.__dict__,
                    derivative_type=derivative_type,
                    derivative_type_description=derivative_type_description,
                    counterparty_name=counterparty_name,
                    counterparty_lei=counterparty_lei,
                    currency_amount_sold=currency_amount_sold,
                    currency_sold=currency_sold,
                    currency_amount_purchased=currency_amount_purchased,
                    currency_purchased=currency_purchased,
                    settlement=settlement,
                    unrealized_appreciation=unrealized_appreciation
                )

        elif derivative_tag.find("futrDeriv"):
            derivative_type = "Future"
            future_tag = derivative_tag.find("futrDeriv")
            base.payoff_profile = child_text(future_tag, "payOffProf")

            reference_instrument_tag = future_tag.find("descRefInstrmnt")
            reference_instrument = None
            if reference_instrument_tag:
                if reference_instrument_tag.find("indexBasketInfo"):
                    reference_instrument = ReferenceIndex.from_xml(reference_instrument_tag.find("indexBasketInfo"))
                elif reference_instrument_tag.find("nestedDerivInfo"):
                    reference_instrument = ReferenceDerivative.from_xml(
                        reference_instrument_tag.find("nestedDerivInfo"))
                else:
                    reference_instrument = ReferenceOther.from_xml(reference_instrument_tag.find("otherRefInst"))

            maturity = child_text(future_tag, "expDate")
            notional = child_text(future_tag, "notionalAmt")
            derivative_currency = child_text(future_tag, "curCd")
            unrealized_appreciation = child_text(future_tag, "unrealizedAppr")
            return ForwardFuture(
                **base.__dict__,
                derivative_type=derivative_type,
                derivative_type_description=derivative_type_description,
                counterparty_name=counterparty_name,
                counterparty_lei=counterparty_lei,
                reference_instrument=reference_instrument,
                maturity=maturity,
                notional=notional,
                derivative_currency=derivative_currency,
                unrealized_appreciation=unrealized_appreciation
            )

        elif derivative_tag.find("swapDeriv"):
            swap_tag = derivative_tag.find("swapDeriv")

            if derivative_tag.find("swapFlag"):
                derivative_type = "Swap"
                swap_flag = child_text(swap_tag, "swapFlag") == "Y"
                maturity = child_text(swap_tag, "terminationDt")

                reference_instrument_tag = swap_tag.find("descRefInstrmnt")
                reference_instrument = None
                if reference_instrument_tag:
                    if reference_instrument_tag.find("indexBasketInfo"):
                        reference_instrument = ReferenceIndex.from_xml(reference_instrument_tag.find("indexBasketInfo"))
                    elif reference_instrument_tag.find("nestedDerivInfo"):
                        reference_instrument = ReferenceDerivative.from_xml(
                            reference_instrument_tag.find("nestedDerivInfo"))
                    else:
                        reference_instrument = ReferenceOther.from_xml(reference_instrument_tag.find("otherRefInst"))

                payment_index = None
                payment_rate = None
                payment_description = None
                payment_currency = None
                payment_amount = None
                if swap_tag.find("fixedPmntDesc"):
                    payment_type = child_value(swap_tag, "fixedPmntDesc", "fixedOrFloating")
                    payment_rate = child_value(swap_tag, "fixedPmntDesc", "fixedRt")
                    payment_currency = child_value(swap_tag, "fixedPmntDesc", "curCd")
                    payment_amount = child_value(swap_tag, "fixedPmntDesc", "amount")
                elif swap_tag.find("floatingPmntDesc"):
                    payment_type = child_value(swap_tag, "floatingPmntDesc", "fixedOrFloating")
                    payment_rate = child_value(swap_tag, "floatingPmntDesc", "floatingRtSpread")
                    payment_index = child_value(swap_tag, "floatingPmntDesc", "floatingRtIndex")
                    payment_currency = child_value(swap_tag, "floatingPmntDesc", "curCd")
                    payment_amount = child_value(swap_tag, "floatingPmntDesc", "pmntAmt")
                elif swap_tag.find("otherPmntDesc"):
                    payment_type = child_value(swap_tag, "otherPmntDesc", "fixedOrFloating")
                    payment_description = child_text(swap_tag, "otherPmntDesc")

                receipt_index = None
                receipt_rate = None
                receipt_description = None
                receipt_currency = None
                receipt_amount = None
                if swap_tag.find("fixedRecDesc"):
                    receipt_type = child_value(swap_tag, "fixedRecDesc", "fixedOrFloating")
                    receipt_rate = child_value(swap_tag, "fixedRecDesc", "fixedRt")
                    receipt_currency = child_value(swap_tag, "fixedRecDesc", "curCd")
                    receipt_amount = child_value(swap_tag, "fixedRecDesc", "amount")
                elif swap_tag.find("floatingRecDesc"):
                    receipt_type = child_value(swap_tag, "floatingRecDesc", "fixedOrFloating")
                    receipt_rate = child_value(swap_tag, "floatingRecDesc", "floatingRtSpread")
                    receipt_index = child_value(swap_tag, "floatingRecDesc", "floatingRtIndex")
                    receipt_currency = child_value(swap_tag, "floatingRecDesc", "curCd")
                    receipt_amount = child_value(swap_tag, "floatingRecDesc", "pmntAmt")
                elif swap_tag.find("otherRecDesc"):
                    receipt_type = child_value(swap_tag, "otherRecDesc", "fixedOrFloating")
                    receipt_description = child_text(swap_tag, "otherRecDesc")

                upfront_payment = child_text(swap_tag, "upfrontPmnt")
                upfront_payment_currency = child_text(swap_tag, "pmntCurCd")
                upfront_receipts = child_text(swap_tag, "upfrontRcpt")
                upfront_receipts_currency = child_text(swap_tag, "rcptCurCd")
                notional = child_text(swap_tag, "notionalAmt")
                derivative_currency = child_text(swap_tag, "curCd")
                unrealized_appreciation = child_text(swap_tag, "unrealizedAppr")
                return Swap(
                    **base.__dict__,
                    derivative_type=derivative_type,
                    derivative_type_description=derivative_type_description,
                    counterparty_name=counterparty_name,
                    counterparty_lei=counterparty_lei,
                    reference_instrument=reference_instrument,
                    swap_flag=swap_flag,
                    maturity=maturity,
                    payment_type=payment_type,
                    payment_description=payment_description,
                    payment_rate=payment_rate,
                    payment_index=payment_index,
                    payment_currency=payment_currency,
                    payment_amount=payment_amount,
                    receipt_type=receipt_type,
                    receipt_description=receipt_description,
                    receipt_rate=receipt_rate,
                    receipt_index=receipt_index,
                    receipt_currency=receipt_currency,
                    receipt_amount=receipt_amount,
                    upfront_payment=upfront_payment,
                    upfront_payment_currency=upfront_payment_currency,
                    upfront_receipts=upfront_receipts,
                    upfront_receipts_currency=upfront_receipts_currency,
                    notional=notional,
                    derivative_currency=derivative_currency,
                    unrealized_appreciation=unrealized_appreciation
                )
            else:
                derivative_type = "FX Swap"
                currency_amount_sold = child_text(swap_tag, "amtCurSold")
                currency_sold = child_text(swap_tag, "curSold")
                currency_amount_purchased = child_text(swap_tag, "amtCurPur")
                currency_purchased = child_text(swap_tag, "curPur")
                settlement = child_text(swap_tag, "settlementDt")
                unrealized_appreciation = child_text(swap_tag, "unrealizedAppr")
                return FXForwardOrSwap(
                    **base.__dict__,
                    derivative_type=derivative_type,
                    derivative_type_description=derivative_type_description,
                    counterparty_name=counterparty_name,
                    counterparty_lei=counterparty_lei,
                    currency_amount_sold=currency_amount_sold,
                    currency_sold=currency_sold,
                    currency_amount_purchased=currency_amount_purchased,
                    currency_purchased=currency_purchased,
                    settlement=settlement,
                    unrealized_appreciation=unrealized_appreciation
                )

        elif derivative_tag.find("optionSwaptionWarrantDeriv"):
            option_tag = derivative_tag.find("optionSwaptionWarrantDeriv")
            derivative_type = {"OPT": "Option", "SWO": "Swaption", "WAR": "Warrant"}.get(
                child_value(derivative_tag, "optionSwaptionWarrantDeriv", "derivCat")
            )
            option_type = child_text(option_tag, "putOrCall")
            option_payoff = child_text(option_tag, "writtenOrPur")

            reference_instrument_tag = option_tag.find("descRefInstrmnt")
            reference_instrument = None
            if reference_instrument_tag:
                if reference_instrument_tag.find("indexBasketInfo"):
                    reference_instrument = ReferenceIndex.from_xml(reference_instrument_tag.find("indexBasketInfo"))
                elif reference_instrument_tag.find("nestedDerivInfo"):
                    reference_instrument = ReferenceDerivative.from_xml(
                        reference_instrument_tag.find("nestedDerivInfo"))
                else:
                    reference_instrument = ReferenceOther.from_xml(reference_instrument_tag.find("otherRefInst"))

            shares = child_text(option_tag, "shareNo")
            notional = child_text(option_tag, "principalAmt")
            derivative_currency = child_text(option_tag, "curCd")
            exercise_price = child_text(option_tag, "exercisePrice")
            exercise_currency = child_text(option_tag, "exercisePriceCurCd")
            exercise_date = child_text(option_tag, "expDt")
            delta = child_text(option_tag, "delta")
            unrealized_appreciation = child_text(option_tag, "unrealizedAppr")
            return Option(
                **base.__dict__,
                derivative_type=derivative_type,
                derivative_type_description=derivative_type_description,
                counterparty_name=counterparty_name,
                counterparty_lei=counterparty_lei,
                option_type=option_type,
                option_payoff=option_payoff,
                reference_instrument=reference_instrument,
                shares=shares,
                notional=notional,
                derivative_currency=derivative_currency,
                exercise_price=exercise_price,
                exercise_currency=exercise_currency,
                exercise_date=exercise_date,
                delta=delta,
                unrealized_appreciation=unrealized_appreciation
            )

        elif derivative_tag.find("othDeriv"):
            derivative_type = "Other"
            derivative_type_description = child_value(derivative_tag, "othDeriv", "othDesc")
            other_derivative_tag = derivative_tag.find("othDeriv")

            reference_instrument_tag = other_derivative_tag.find("descRefInstrmnt")
            reference_instrument = None
            if reference_instrument_tag:
                if reference_instrument_tag.find("indexBasketInfo"):
                    reference_instrument = ReferenceIndex.from_xml(reference_instrument_tag.find("indexBasketInfo"))
                elif reference_instrument_tag.find("nestedDerivInfo"):
                    reference_instrument = ReferenceDerivative.from_xml(reference_instrument_tag.find("nestedDerivInfo"))
                else:
                    reference_instrument = ReferenceOther.from_xml(reference_instrument_tag.find("otherRefInst"))

            maturity = child_text(other_derivative_tag, "terminationDt")
            notional_tag = other_derivative_tag.find("notionalAmts")
            notional = [el.attrs.get("amt") for el in notional_tag.find_all("notionalAmt")]
            notional_currency = [el.attrs.get("curCd") for el in notional_tag.find_all("notionalAmt")]
            delta = child_text(other_derivative_tag, "delta")
            unrealized_appreciation = child_text(other_derivative_tag, "unrealizedAppr")
            return OtherDerivative(
                **base.__dict__,
                derivative_type=derivative_type,
                derivative_type_description=derivative_type_description,
                counterparty_name=counterparty_name,
                counterparty_lei=counterparty_lei,
                reference_instrument=reference_instrument,
                maturity=maturity,
                notional=notional,
                notional_currency=notional_currency,
                delta=delta,
                unrealized_appreciation=unrealized_appreciation
            )


class ForwardFuture(Derivative):
    reference_instrument: Optional[object]
    maturity: Optional[date]
    notional: Optional[Decimal]
    derivative_currency: Optional[str]
    unrealized_appreciation: Optional[Decimal]


class Swap(Derivative):
    reference_instrument: Optional[object]
    swap_flag: bool
    maturity: date
    payment_type: str
    payment_description: Optional[str]
    payment_rate: Optional[Decimal]
    payment_index: Optional[str]
    payment_currency: Optional[str]
    payment_amount: Optional[Decimal]
    receipt_type: str
    receipt_description: Optional[str]
    receipt_rate: Optional[Decimal]
    receipt_index: Optional[str]
    receipt_currency: Optional[str]
    receipt_amount: Optional[Decimal]
    upfront_payment: Decimal
    upfront_payment_currency: str
    upfront_receipts: Decimal
    upfront_receipts_currency: str
    notional: Decimal
    derivative_currency: str
    unrealized_appreciation: Optional[Decimal]


class FXForwardOrSwap(Derivative):
    currency_amount_sold: Optional[Decimal]
    currency_sold: Optional[str]
    currency_amount_purchased: Optional[Decimal]
    currency_purchased: Optional[str]
    settlement: Optional[date]
    unrealized_appreciation: Optional[Decimal]


class Option(Derivative):
    option_type: str
    option_payoff: str
    reference_instrument: Optional[object]
    shares: Optional[Decimal]
    notional: Optional[Decimal]
    derivative_currency: Optional[str]
    exercise_price: Decimal
    exercise_currency: str
    exercise_date: date
    delta: Optional[Decimal]
    unrealized_appreciation: Optional[Decimal]


class OtherDerivative(Derivative):
    reference_instrument: Optional[object]
    maturity: date
    notional: list[Decimal]
    notional_currency: list[str]
    delta: Optional[Decimal]
    unrealized_appreciation: Optional[Decimal]


class ReferenceDerivative(BaseModel):
    derivative_type: str
    derivative_type_description: Optional[str]
    counterparty_name: Optional[str]
    counterparty_lei: Optional[str]

    @classmethod
    def from_xml(cls, tag):
        counterparty_tag = tag.find("counterparties")
        counterparty_name = child_text(counterparty_tag, "counterpartyName")
        counterparty_lei = child_text(counterparty_tag, "counterpartyLei")

        derivative_type_description = None
        if tag.find("fwdDeriv"):
            forward_tag = tag.find("fwdDeriv")

            if forward_tag.find("payOffProf"):
                derivative_type = "Forward"
                payoff_profile = child_text(forward_tag, "payOffProf")

                reference_instrument_tag = forward_tag.find("descRefInstrmnt")
                reference_instrument = None
                if reference_instrument_tag:
                    if reference_instrument_tag.find("indexBasketInfo"):
                        reference_instrument = ReferenceIndex.from_xml(reference_instrument_tag.find("indexBasketInfo"))
                    elif reference_instrument_tag.find("nestedDerivInfo"):
                        reference_instrument = ReferenceDerivative.from_xml(reference_instrument_tag.find("nestedDerivInfo"))
                    else:
                        reference_instrument = ReferenceOther.from_xml(reference_instrument_tag.find("otherRefInst"))

                maturity = child_text(forward_tag, "expDate")
                notional = child_text(forward_tag, "notionalAmt")
                derivative_currency = child_text(forward_tag, "curCd")
                return ReferenceForwardFuture(
                    derivative_type=derivative_type,
                    derivative_type_description=derivative_type_description,
                    counterparty_name=counterparty_name,
                    counterparty_lei=counterparty_lei,
                    payoff_profile=payoff_profile,
                    reference_instrument=reference_instrument,
                    maturity=maturity,
                    notional=notional,
                    derivative_currency=derivative_currency
                )

            else:
                derivative_type = "FX Forward"
                currency_amount_sold = child_text(forward_tag, "amtCurSold")
                currency_sold = child_text(forward_tag, "curSold")
                currency_amount_purchased = child_text(forward_tag, "amtCurPur")
                currency_purchased = child_text(forward_tag, "curPur")
                settlement = child_text(forward_tag, "settlementDt")
                return ReferenceFXForwardOrSwap(
                    derivative_type=derivative_type,
                    derivative_type_description=derivative_type_description,
                    counterparty_name=counterparty_name,
                    counterparty_lei=counterparty_lei,
                    currency_amount_sold=currency_amount_sold,
                    currency_sold=currency_sold,
                    currency_amount_purchased=currency_amount_purchased,
                    currency_purchased=currency_purchased,
                    settlement=settlement
                )

        elif tag.find("futrDeriv"):
            derivative_type = "Future"
            future_tag = tag.find("futrDeriv")
            payoff_profile = child_text(future_tag, "payOffProf")

            reference_instrument_tag = future_tag.find("descRefInstrmnt")
            reference_instrument = None
            if reference_instrument_tag:
                if reference_instrument_tag.find("indexBasketInfo"):
                    reference_instrument = ReferenceIndex.from_xml(reference_instrument_tag.find("indexBasketInfo"))
                elif reference_instrument_tag.find("nestedDerivInfo"):
                    reference_instrument = ReferenceDerivative.from_xml(
                        reference_instrument_tag.find("nestedDerivInfo"))
                else:
                    reference_instrument = ReferenceOther.from_xml(reference_instrument_tag.find("otherRefInst"))

            maturity = child_text(future_tag, "expDate")
            notional = child_text(future_tag, "notionalAmt")
            derivative_currency = child_text(future_tag, "curCd")
            return ReferenceForwardFuture(
                derivative_type=derivative_type,
                derivative_type_description=derivative_type_description,
                counterparty_name=counterparty_name,
                counterparty_lei=counterparty_lei,
                payoff_profile=payoff_profile,
                reference_instrument=reference_instrument,
                maturity=maturity,
                notional=notional,
                derivative_currency=derivative_currency
            )

        elif tag.find("swapDeriv"):
            swap_tag = tag.find("swapDeriv")

            if tag.find("swapFlag"):
                derivative_type = "Swap"
                swap_flag = child_text(swap_tag, "swapFlag") == "Y"
                maturity = child_text(swap_tag, "terminationDt")

                reference_instrument_tag = swap_tag.find("descRefInstrmnt")
                reference_instrument = None
                if reference_instrument_tag:
                    if reference_instrument_tag.find("indexBasketInfo"):
                        reference_instrument = ReferenceIndex.from_xml(reference_instrument_tag.find("indexBasketInfo"))
                    elif reference_instrument_tag.find("nestedDerivInfo"):
                        reference_instrument = ReferenceDerivative.from_xml(
                            reference_instrument_tag.find("nestedDerivInfo"))
                    else:
                        reference_instrument = ReferenceOther.from_xml(reference_instrument_tag.find("otherRefInst"))

                payment_index = None
                payment_rate = None
                payment_description = None
                payment_currency = None
                payment_amount = None
                if swap_tag.find("fixedPmntDesc"):
                    payment_type = child_value(swap_tag, "fixedPmntDesc", "fixedOrFloating")
                    payment_rate = child_value(swap_tag, "fixedPmntDesc", "fixedRt")
                    payment_currency = child_value(swap_tag, "fixedPmntDesc", "curCd")
                    payment_amount = child_value(swap_tag, "fixedPmntDesc", "amount")
                elif swap_tag.find("floatingPmntDesc"):
                    payment_type = child_value(swap_tag, "floatingPmntDesc", "fixedOrFloating")
                    payment_rate = child_value(swap_tag, "floatingPmntDesc", "floatingRtSpread")
                    payment_index = child_value(swap_tag, "floatingPmntDesc", "floatingRtIndex")
                    payment_currency = child_value(swap_tag, "floatingPmntDesc", "curCd")
                    payment_amount = child_value(swap_tag, "floatingPmntDesc", "pmntAmt")
                elif swap_tag.find("otherPmntDesc"):
                    payment_type = child_value(swap_tag, "otherPmntDesc", "fixedOrFloating")
                    payment_description = child_text(swap_tag, "otherPmntDesc")

                receipt_index = None
                receipt_rate = None
                receipt_description = None
                receipt_currency = None
                receipt_amount = None
                if swap_tag.find("fixedRecDesc"):
                    receipt_type = child_value(swap_tag, "fixedRecDesc", "fixedOrFloating")
                    receipt_rate = child_value(swap_tag, "fixedRecDesc", "fixedRt")
                    receipt_currency = child_value(swap_tag, "fixedRecDesc", "curCd")
                    receipt_amount = child_value(swap_tag, "fixedRecDesc", "amount")
                elif swap_tag.find("floatingRecDesc"):
                    receipt_type = child_value(swap_tag, "floatingRecDesc", "fixedOrFloating")
                    receipt_rate = child_value(swap_tag, "floatingRecDesc", "floatingRtSpread")
                    receipt_index = child_value(swap_tag, "floatingRecDesc", "floatingRtIndex")
                    receipt_currency = child_value(swap_tag, "floatingRecDesc", "curCd")
                    receipt_amount = child_value(swap_tag, "floatingRecDesc", "pmntAmt")
                elif swap_tag.find("otherRecDesc"):
                    receipt_type = child_value(swap_tag, "otherRecDesc", "fixedOrFloating")
                    receipt_description = child_text(swap_tag, "otherRecDesc")

                upfront_payment = child_text(swap_tag, "upfrontPmnt")
                upfront_payment_currency = child_text(swap_tag, "pmntCurCd")
                upfront_receipts = child_text(swap_tag, "upfrontRcpt")
                upfront_receipts_currency = child_text(swap_tag, "rcptCurCd")
                notional = child_text(swap_tag, "notionalAmt")
                derivative_currency = child_text(swap_tag, "curCd")
                return ReferenceSwap(
                    derivative_type=derivative_type,
                    derivative_type_description=derivative_type_description,
                    counterparty_name=counterparty_name,
                    counterparty_lei=counterparty_lei,
                    reference_instrument=reference_instrument,
                    swap_flag=swap_flag,
                    maturity=maturity,
                    payment_type=payment_type,
                    payment_description=payment_description,
                    payment_rate=payment_rate,
                    payment_index=payment_index,
                    payment_currency=payment_currency,
                    payment_amount=payment_amount,
                    receipt_type=receipt_type,
                    receipt_description=receipt_description,
                    receipt_rate=receipt_rate,
                    receipt_index=receipt_index,
                    receipt_currency=receipt_currency,
                    receipt_amount=receipt_amount,
                    upfront_payment=upfront_payment,
                    upfront_payment_currency=upfront_payment_currency,
                    upfront_receipts=upfront_receipts,
                    upfront_receipts_currency=upfront_receipts_currency,
                    notional=notional,
                    derivative_currency=derivative_currency
                )
            else:
                derivative_type = "FX Swap"
                currency_amount_sold = child_text(swap_tag, "amtCurSold")
                currency_sold = child_text(swap_tag, "curSold")
                currency_amount_purchased = child_text(swap_tag, "amtCurPur")
                currency_purchased = child_text(swap_tag, "curPur")
                settlement = child_text(swap_tag, "settlementDt")
                return ReferenceFXForwardOrSwap(
                    derivative_type=derivative_type,
                    derivative_type_description=derivative_type_description,
                    counterparty_name=counterparty_name,
                    counterparty_lei=counterparty_lei,
                    currency_amount_sold=currency_amount_sold,
                    currency_sold=currency_sold,
                    currency_amount_purchased=currency_amount_purchased,
                    currency_purchased=currency_purchased,
                    settlement=settlement
                )

        elif tag.find("optionSwaptionWarrantDeriv"):
            option_tag = tag.find("optionSwaptionWarrantDeriv")
            derivative_type = {"OPT": "Option", "SWO": "Swaption", "WAR": "Warrant"}.get(
                child_value(tag, "optionSwaptionWarrantDeriv", "derivCat")
            )
            option_type = child_text(option_tag, "putOrCall")
            option_payoff = child_text(option_tag, "writtenOrPur")

            reference_instrument_tag = option_tag.find("descRefInstrmnt")
            reference_instrument = None
            if reference_instrument_tag:
                if reference_instrument_tag.find("indexBasketInfo"):
                    reference_instrument = ReferenceIndex.from_xml(reference_instrument_tag.find("indexBasketInfo"))
                elif reference_instrument_tag.find("nestedDerivInfo"):
                    reference_instrument = ReferenceDerivative.from_xml(
                        reference_instrument_tag.find("nestedDerivInfo"))
                else:
                    reference_instrument = ReferenceOther.from_xml(reference_instrument_tag.find("otherRefInst"))

            shares = child_text(option_tag, "shareNo")
            notional = child_text(option_tag, "principalAmt")
            derivative_currency = child_text(option_tag, "curCd")
            exercise_price = child_text(option_tag, "exercisePrice")
            exercise_currency = child_text(option_tag, "exercisePriceCurCd")
            exercise_date = child_text(option_tag, "expDt")
            delta = child_text(option_tag, "delta")
            return ReferenceOption(
                derivative_type=derivative_type,
                derivative_type_description=derivative_type_description,
                counterparty_name=counterparty_name,
                counterparty_lei=counterparty_lei,
                option_type=option_type,
                option_payoff=option_payoff,
                reference_instrument=reference_instrument,
                shares=shares,
                notional=notional,
                derivative_currency=derivative_currency,
                exercise_price=exercise_price,
                exercise_currency=exercise_currency,
                exercise_date=exercise_date,
                delta=delta
            )

        elif tag.find("othDeriv"):
            derivative_type = "Other"
            derivative_type_description = child_value(tag, "othDeriv", "othDesc")
            other_derivative_tag = tag.find("othDeriv")

            reference_instrument_tag = other_derivative_tag.find("descRefInstrmnt")
            reference_instrument = None
            if reference_instrument_tag:
                if reference_instrument_tag.find("indexBasketInfo"):
                    reference_instrument = ReferenceIndex.from_xml(reference_instrument_tag.find("indexBasketInfo"))
                elif reference_instrument_tag.find("nestedDerivInfo"):
                    reference_instrument = ReferenceDerivative.from_xml(
                        reference_instrument_tag.find("nestedDerivInfo"))
                else:
                    reference_instrument = ReferenceOther.from_xml(reference_instrument_tag.find("otherRefInst"))

            maturity = child_text(other_derivative_tag, "terminationDt")
            notional_tag = other_derivative_tag.find("notionalAmts")
            notional = [el.attrs.get("amt") for el in notional_tag.find_all("notionalAmt")]
            notional_currency = [el.attrs.get("curCd") for el in notional_tag.find_all("notionalAmt")]
            delta = child_text(other_derivative_tag, "delta")
            return ReferenceOtherDerivative(
                derivative_type=derivative_type,
                derivative_type_description=derivative_type_description,
                counterparty_name=counterparty_name,
                counterparty_lei=counterparty_lei,
                reference_instrument=reference_instrument,
                maturity=maturity,
                notional=notional,
                notional_currency=notional_currency,
                delta=delta,
                unrealized_appreciation=None
            )


class ReferenceForwardFuture(ReferenceDerivative):
    derivativepayoff_profile: Optional[str]
    reference_instrument: Optional[object]
    maturity: Optional[date]
    notional: Optional[Decimal]
    derivative_currency: Optional[str]


class ReferenceSwap(ReferenceDerivative):
    reference_instrument: Optional[object]
    swap_flag: bool
    maturity: date
    payment_type: str
    payment_description: Optional[str]
    payment_rate: Optional[Decimal]
    payment_index: Optional[str]
    payment_currency: Optional[str]
    payment_amount: Optional[Decimal]
    receipt_type: str
    receipt_description: Optional[str]
    receipt_rate: Optional[Decimal]
    receipt_index: Optional[str]
    receipt_currency: Optional[str]
    receipt_amount: Optional[Decimal]
    upfront_payment: Decimal
    upfront_payment_currency: str
    upfront_receipts: Decimal
    upfront_receipts_currency: str
    notional: Decimal
    derivative_currency: str


class ReferenceFXForwardOrSwap(ReferenceDerivative):
    currency_amount_sold: Optional[Decimal]
    currency_sold: Optional[str]
    currency_amount_purchased: Optional[Decimal]
    currency_purchased: Optional[str]
    settlement: Optional[date]


class ReferenceOption(ReferenceDerivative):
    option_type: str
    option_payoff: str
    reference_instrument: Optional[object]
    shares: Optional[Decimal]
    notional: Optional[Decimal]
    derivative_currency: Optional[str]
    exercise_price: Decimal
    exercise_currency: str
    exercise_date: date
    delta: Optional[Decimal]


class ReferenceOtherDerivative(ReferenceDerivative):
    reference_instrument: Optional[object]
    maturity: date
    notional: list[Decimal]
    notional_currency: list[str]
    delta: Optional[Decimal]


class ReferenceIndexComponents(BaseModel):
    name: str
    cusip: Optional[str]
    isin: Optional[str]
    ticker: Optional[str]
    other_identifier_name: Optional[str]
    other_identifier: Optional[str]
    balance: Decimal
    currency: Optional[str]
    issue_currency: str
    value: Decimal

    @classmethod
    def from_xml(cls, tag):
        name = child_text(tag, "othIndName")
        identifiers_tag = tag.find("identifiers")
        cusip = child_text(identifiers_tag, "cusip")
        isin = child_text(identifiers_tag, "isin")
        ticker = child_text(identifiers_tag, "ticker")
        other_identifier_name = child_value(identifiers_tag, "other", "otherDesc")
        other_identifier = child_value(identifiers_tag, "other", "value")
        balance = child_text(tag, "othIndNotAmt")
        currency = child_text(tag, "othIndCurCd")
        value = child_text(tag, "othIndValue")
        issue_currency = child_text(tag, "othIndIssCurCd")
        if currency is None:
            currency = issue_currency
        return cls(
            name=name,
            cusip=cusip,
            isin=isin,
            ticker=ticker,
            other_identifier_name=other_identifier_name,
            other_identifier=other_identifier,
            balance=balance,
            currency=currency,
            value=value,
            issue_currency=issue_currency
        )


class ReferenceIndex(BaseModel):
    name: Optional[str]
    identifier: Optional[str]
    description: Optional[str]
    components: Optional[list[ReferenceIndexComponents]]

    @classmethod
    def from_xml(cls, tag):
        name = child_text(tag, "indexName")
        identifier = child_text(tag, "indexIdentifier")
        description = child_text(tag, "narrativeDesc")
        components = None
        component_tag = tag.find("components")
        if component_tag:
            components = [ReferenceIndexComponents.from_xml(instrument_tag) for instrument_tag in component_tag.find_all("component")]
        return cls(
            name=name,
            identifier=identifier,
            description=description,
            components=components
        )


class ReferenceOther(BaseModel):
    issuer: Optional[str]
    name: Optional[str]
    cusip: Optional[str]
    isin: Optional[str]
    ticker: Optional[str]
    other_identifier_name: Optional[str]
    other_identifier: Optional[str]

    @classmethod
    def from_xml(cls, tag):
        issuer = child_text(tag, "issuerName")
        name = child_text(tag, "issueTitle")
        identifier_tag = tag.find("identifiers")
        cusip = child_text(identifier_tag, "cusip")
        isin = child_text(identifier_tag, "isin")
        ticker = child_text(identifier_tag, "ticker")
        other_identifier_name = child_value(identifier_tag, "other", "otherDesc")
        other_identifier = child_value(identifier_tag, "other")
        return cls(
            issuer=issuer,
            name=name,
            cusip=cusip,
            isin=isin,
            ticker=ticker,
            other_identifier_name=other_identifier_name,
            other_identifier=other_identifier
        )



import gzip
import os
import time
import zipfile
from collections import deque
from functools import wraps
from io import BytesIO
from pathlib import Path
from threading import Lock
from typing import Union, Optional

import httpx
import orjson as json
from stamina import retry
from tqdm import tqdm

__all__ = ["get_with_retry", "get_with_retry_async", "stream_with_retry", "post_with_retry", "post_with_retry_async",
           "download_file", "download_file_async", "download_json", "download_json_async", "stream_file",
           "download_text", "download_text_between_tags", "download_bulk_data", "download_datafile"]

attempts = 6
retry_timeout = 40
wait_initial = 0.1
max_requests_per_second = 8
http_timeout = 20
max_connections = 5
text_extensions = (".txt", ".htm", ".html", ".xsd", ".xml", "XML", ".json", ".idx", ".paper")

default_local_data_dir = Path(os.path.join(os.path.expanduser("~"), ".edgar"))
edgar_data_dir = Path(os.getenv('EDGAR_LOCAL_DATA_DIR', default_local_data_dir))
if not edgar_data_dir.exists():
    os.makedirs(edgar_data_dir)


class TooManyRequestsError(Exception):
    def __init__(self, url, message="Too Many Requests"):
        self.url = url
        self.message = message
        super().__init__(self.message)


class IdentityNotSetException(Exception):
    pass


class RequestRate:
    """
    A simple class to represent a request rate, i.e., the maximum number of requests for a given time window
    """

    def __init__(self, max_requests: int, time_window: int):
        if max_requests <= 0:
            raise ValueError("max_requests must be a positive integer")
        if time_window <= 0:
            raise ValueError("time_window must be a positive integer")

        self.max_requests: int = max_requests
        self.time_window: int = time_window


class Throttler:
    """
    A simple throttler that limits the number of requests per time window
    """

    def __init__(self, request_rate: RequestRate, sleep_interval=0.1):
        self.request_rate = request_rate
        self.sleep_interval = sleep_interval
        self.request_timestamps = deque()
        self.lock = Lock()
        self.decorated_functions = None
        self.total_calls = 0
        self.peak_call_rate: float = 0.0

    def get_ticket(self):
        with self.lock:
            current_time = time.monotonic()

            # Remove timestamps older than the time window
            while self.request_timestamps and self.request_timestamps[
                0] <= current_time - self.request_rate.time_window:
                self.request_timestamps.popleft()

            if len(self.request_timestamps) < self.request_rate.max_requests:
                self.request_timestamps.append(current_time)
                return True
            else:
                return False

    def wait_for_ticket(self):
        while not self.get_ticket():
            time.sleep(self.sleep_interval)

    def update_metrics(self):
        self.total_calls += 1
        current_call_rate: float = len(self.request_timestamps) / self.request_rate.time_window
        self.peak_call_rate = max(self.peak_call_rate, current_call_rate)

    def get_metrics(self):
        return {
            "decorated_functions": self.decorated_functions,  # Now it's a list
            "total_calls": self.total_calls,
            "peak_call_rate": self.peak_call_rate,
            "request_rate_limit": self.request_rate.max_requests,
        }

    def print_metrics(self):
        metrics = self.get_metrics()
        print(f"Metrics for decorated functions: {', '.join(metrics['decorated_functions'])}")
        print(f"Total calls: {metrics['total_calls']}")
        print(f"Peak call rate: {metrics['peak_call_rate']:.2f} calls per second")


_throttler_instances = {}  # Singleton instance for throttler


def throttle_requests(request_rate=None, requests_per_second=None, **kwargs):
    """
    Decorator to throttle the number of requests per second.
    """
    if requests_per_second is not None:
        request_rate = RequestRate(max_requests=requests_per_second, time_window=1)
    elif request_rate is None:
        raise ValueError("Either request_rate or requests_per_second must be provided")

    # Use a single key for all instances to create a global throttler
    key = "global_throttler"

    if key not in _throttler_instances:
        _throttler_instances[key] = Throttler(request_rate, **kwargs)

    throttler = _throttler_instances[key]

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            throttler.wait_for_ticket()
            result = func(*args, **kwargs)
            throttler.update_metrics()
            return result

        # Store the decorated function name
        if throttler.decorated_functions is None:
            throttler.decorated_functions = []
        throttler.decorated_functions.append(func.__name__)

        return wrapper

    return decorator


def is_redirect(response):
    return response.status_code in [301, 302]


def with_identity(func):
    @wraps(func)
    def wrapper(url, identity=None, identity_callable=None, *args, **kwargs):
        if identity is None:
            if identity_callable is not None:
                identity = identity_callable()
            else:
                identity = os.environ.get("EDGAR_IDENTITY")
        if identity is None:
            raise IdentityNotSetException("User-Agent identity is not set")

        headers = kwargs.get("headers", {})
        headers["User-Agent"] = identity
        kwargs["headers"] = headers

        return func(url, identity=identity, identity_callable=identity_callable, *args, **kwargs)

    return wrapper


@retry(on=httpx.RequestError, attempts=attempts, timeout=retry_timeout, wait_initial=wait_initial)
@with_identity
@throttle_requests(requests_per_second=max_requests_per_second)
def get_with_retry(url, identity=None, identity_callable=None, **kwargs):
    """
    Sends a GET request with retry functionality and identity handling.

    Args:
        url (str): The URL to send the GET request to.
        identity (str, optional): The identity to use for the request. Defaults to None.
        identity_callable (callable, optional): A callable that returns the identity. Defaults to None.
        **kwargs: Additional keyword arguments to pass to the underlying httpx.Client.get() method.

    Returns:
        httpx.Response: The response object returned by the GET request.

    Raises:
        TooManyRequestsError: If the response status code is 429 (Too Many Requests).
    """
    with httpx.Client(timeout=http_timeout) as client:
        response = client.get(url, **kwargs)
        if response.status_code == 429:
            raise TooManyRequestsError(url)
        elif is_redirect(response):
            return get_with_retry(response.headers["Location"], identity=identity, identity_callable=identity_callable,
                                  **kwargs)
        return response


@retry(on=httpx.RequestError, attempts=attempts, timeout=retry_timeout, wait_initial=wait_initial)
@with_identity
@throttle_requests(requests_per_second=max_requests_per_second)
async def get_with_retry_async(url, identity=None, identity_callable=None, **kwargs):
    """
    Sends an asynchronous GET request with retry functionality and identity handling.

    Args:
        url (str): The URL to send the GET request to.
        identity (str, optional): The identity to use for the request. Defaults to None.
        identity_callable (callable, optional): A callable that returns the identity. Defaults to None.
        **kwargs: Additional keyword arguments to pass to the underlying httpx.AsyncClient.get() method.

    Returns:
        httpx.Response: The response object returned by the GET request.

    Raises:
        TooManyRequestsError: If the response status code is 429 (Too Many Requests).
    """
    async with httpx.AsyncClient(timeout=http_timeout) as client:
        response = await client.get(url, **kwargs)
        if response.status_code == 429:
            raise TooManyRequestsError(url)
        elif is_redirect(response):
            return await get_with_retry_async(response.headers["Location"], identity=identity,
                                              identity_callable=identity_callable, **kwargs)
        return response


@retry(on=httpx.RequestError, attempts=attempts, timeout=retry_timeout, wait_initial=wait_initial)
@with_identity
@throttle_requests(requests_per_second=max_requests_per_second)
def stream_with_retry(url, identity=None, identity_callable=None, **kwargs):
    """
    Sends a streaming GET request with retry functionality and identity handling.

    Args:
        url (str): The URL to send the streaming GET request to.
        identity (str, optional): The identity to use for the request. Defaults to None.
        identity_callable (callable, optional): A callable that returns the identity. Defaults to None.
        **kwargs: Additional keyword arguments to pass to the underlying httpx.Client.stream() method.

    Yields:
        bytes: The bytes of the response content.

    Raises:
        TooManyRequestsError: If the response status code is 429 (Too Many Requests).
    """
    with httpx.Client(timeout=http_timeout) as client:
        with client.stream("GET", url, **kwargs) as response:
            if response.status_code == 429:
                raise TooManyRequestsError(url)
            elif is_redirect(response):
                yield stream_with_retry(response.headers["Location"],
                                        identity=identity,
                                        identity_callable=identity_callable, **kwargs)
            else:
                yield response


@retry(on=httpx.RequestError, attempts=attempts, timeout=retry_timeout, wait_initial=wait_initial)
@with_identity
@throttle_requests(requests_per_second=max_requests_per_second)
def post_with_retry(url, data=None, json=None, identity=None, identity_callable=None,
                    **kwargs):
    """
    Sends a POST request with retry functionality and identity handling.

    Args:
        url (str): The URL to send the POST request to.
        data (dict, optional): The data to include in the request body. Defaults to None.
        json (dict, optional): The JSON data to include in the request body. Defaults to None.
        identity (str, optional): The identity to use for the request. Defaults to None.
        identity_callable (callable, optional): A callable that returns the identity. Defaults to None.
        **kwargs: Additional keyword arguments to pass to the underlying httpx.Client.post() method.

    Returns:
        httpx.Response: The response object returned by the POST request.

    Raises:
        TooManyRequestsError: If the response status code is 429 (Too Many Requests).
    """
    with httpx.Client(timeout=http_timeout) as client:
        response = client.post(url, data=data, json=json, **kwargs)
        if response.status_code == 429:
            raise TooManyRequestsError(url)
        elif is_redirect(response):
            return post_with_retry(response.headers["Location"], data=data, json=json, identity=identity,
                                   identity_callable=identity_callable, **kwargs)
        return response


@retry(on=httpx.RequestError, attempts=attempts, timeout=retry_timeout, wait_initial=wait_initial)
@with_identity
@throttle_requests(requests_per_second=max_requests_per_second)
async def post_with_retry_async(url,
                                data=None,
                                json=None,
                                identity=None,
                                identity_callable=None, **kwargs):
    """
    Sends an asynchronous POST request with retry functionality and identity handling.

    Args:
        url (str): The URL to send the POST request to.
        data (dict, optional): The data to include in the request body. Defaults to None.
        json (dict, optional): The JSON data to include in the request body. Defaults to None.
        identity (str, optional): The identity to use for the request. Defaults to None.
        identity_callable (callable, optional): A callable that returns the identity. Defaults to None.
        **kwargs: Additional keyword arguments to pass to the underlying httpx.AsyncClient.post() method.

    Returns:
        httpx.Response: The response object returned by the POST request.

    Raises:
        TooManyRequestsError: If the response status code is 429 (Too Many Requests).
    """
    async with httpx.AsyncClient(timeout=http_timeout) as client:
        response = await client.post(url, data=data, json=json, **kwargs)
        if response.status_code == 429:
            raise TooManyRequestsError(url)
        elif is_redirect(response):
            return await post_with_retry_async(response.headers["Location"], data=data, json=json, identity=identity,
                                               identity_callable=identity_callable, **kwargs)
        return response


def inspect_response(response: httpx.Response):
    """
    Check if the response is successful and raise an exception if not.
    """
    if response.status_code != 200:
        response.raise_for_status()


def decode_content(content: bytes) -> str:
    """
    Decode the content of a file.
    """
    try:
        return content.decode('utf-8')
    except UnicodeDecodeError:
        return content.decode('latin-1')


def save_or_return_content(content: Union[str, bytes], path: Optional[Union[str, Path]]) -> Union[
    str, bytes, None]:
    """
    Save the content to a specified path or return the content.

    Args:
        content (str or bytes): The content to save or return.
        path (str or Path, optional): The path where the content should be saved. If None, return the content.
        as_text (bool): Whether the content is text or binary.

    Returns:
        str or bytes or None: The content if not saved, or None if saved.
    """
    if path is not None:
        path = Path(path)

        # Determine if the path is a directory or a file
        if path.is_dir():
            file_name = "downloaded_file"  # Replace with logic to extract file name from URL if available
            file_path = path / file_name
        else:
            file_path = path

        # Save the file
        if isinstance(content, bytes):
            file_path.write_bytes(content)
        else:
            file_path.write_text(content)

        return None

    return content


def download_file(url: str, as_text: bool = None, path: Optional[Union[str, Path]] = None) -> Union[str, bytes, None]:
    """
    Download a file from a URL.

    Args:
        url (str): The URL of the file to download.
        as_text (bool, optional): Whether to download the file as text or binary.
        path (str or Path, optional): The path where the file should be saved.
        If None, the default is determined based on the file extension. Defaults to None.

    Returns:
        str or bytes: The content of the downloaded file, either as text or binary data.
    """
    if as_text is None:
        # Set the default based on the file extension
        as_text = url.endswith(text_extensions)

    response = get_with_retry(url=url)
    inspect_response(response)

    if not as_text:
        # Set the default to true if the url ends with a text extension
        as_text = any([url.endswith(ext) for ext in text_extensions])

    # Check if the content is gzip-compressed
    if url.endswith("gz"):
        binary_file = BytesIO(response.content)
        with gzip.open(binary_file, 'rb') as f:
            file_content = f.read()
            if as_text:
                file_content = decode_content(file_content)
    else:
        # If we explicitly asked for text or there is an encoding, try to return text
        if as_text:
            file_content = response.text
            # Should get here for jpg and PDFs
        else:
            file_content = response.content

    if path and path.is_dir():
        path = path / os.path.basename(url)
    return save_or_return_content(file_content, path)


async def download_file_async(url: str, as_text: bool = None, path: Optional[Union[str, Path]] = None) -> Union[
    str, bytes, None]:
    """
    Download a file from a URL asynchronously.

    Args:
        url (str): The URL of the file to download.
        as_text (bool, optional): Whether to download the file as text or binary.
            If None, the default is determined based on the file extension. Defaults to None.
        path (str or Path, optional): The path where the file should be saved.

    Returns:
        str or bytes: The content of the downloaded file, either as text or binary data.
    """
    if as_text is None:
        # Set the default based on the file extension
        as_text = url.endswith(text_extensions)

    response = await get_with_retry_async(url)
    inspect_response(response)

    if as_text:
        # Download as text
        return response.text
    else:
        # Download as binary
        content = response.content

        # Check if the content is gzip-compressed
        if response.headers.get("Content-Encoding") == "gzip":
            content = gzip.decompress(content)

    if path and path.is_dir():
        path = path / os.path.basename(url)

    return save_or_return_content(content, path)


CHUNK_SIZE = 4 * 1024 * 1024  # 4MB


@retry(on=httpx.RequestError, attempts=attempts, timeout=retry_timeout, wait_initial=wait_initial)
@with_identity
@throttle_requests(requests_per_second=max_requests_per_second)
async def stream_file(url: str, as_text: bool = None, path: Optional[Union[str, Path]] = None, **kwargs) -> Union[
    str, bytes, None]:
    """
    Download a file from a URL asynchronously with progress bar using httpx.

    Args:
        url (str): The URL of the file to download.
        as_text (bool, optional): Whether to download the file as text or binary.
            If None, the default is determined based on the file extension. Defaults to None.
        path (str or Path, optional): The path where the file should be saved.

    Returns:
        str or bytes: The content of the downloaded file, either as text or binary data.
    """
    if as_text is None:
        # Set the default based on the file extension
        as_text = url.endswith(text_extensions)

    async with httpx.AsyncClient(timeout=http_timeout, headers=kwargs['headers']) as client:
        async with client.stream('GET', url) as response:
            inspect_response(response)
            total_size = int(response.headers.get('Content-Length', 0))

            if as_text:
                # Download as text
                content = await response.text()
                return content
            else:
                # Download as binary
                content = b''
                progress_bar = tqdm(
                    total=total_size / (1024 * 1024),  # Convert to MB
                    unit='MB',
                    unit_scale=True,
                    unit_divisor=1024,
                    bar_format='{l_bar}{bar}| {n:.2f}/{total:.2f}MB [{elapsed}<{remaining}, {rate_fmt}]',
                    desc=f"Downloading {os.path.basename(url)}",
                    ascii=False  # Use Unicode characters for a nicer looking bar
                )
                downloaded = 0
                async for chunk in response.aiter_bytes(chunk_size=CHUNK_SIZE):
                    content += chunk
                    downloaded += len(chunk)
                    progress_bar.update(len(chunk) / (1024 * 1024))  # Update in MB
                progress_bar.close()

                # Check if the content is gzip-compressed
                if response.headers.get("Content-Encoding") == "gzip":
                    content = gzip.decompress(content)

            if path:
                if isinstance(path, str):
                    path = Path(path)
                if path.is_dir():
                    path = path / os.path.basename(url)

            return save_or_return_content(content, path)


def download_json(data_url: str) -> dict:
    """
    Download JSON data from a URL.

    Args:
        data_url (str): The URL of the JSON data to download.

    Returns:
        dict: The parsed JSON data.
    """
    content = download_file(data_url, as_text=True)
    return json.loads(content)


def download_text(url: str) -> Optional[str]:
    return download_file(url, as_text=True)


async def download_json_async(data_url: str) -> dict:
    """
    Download JSON data from a URL asynchronously.

    Args:
        data_url (str): The URL of the JSON data to download.

    Returns:
        dict: The parsed JSON data.
    """
    content = await download_file_async(data_url, as_text=True)
    return json.loads(content)


def download_text_between_tags(url: str, tag: str):
    """
    Download the content of a URL and extract the text between the tags
    This is mainly for reading the header of a filing

    :param url: The URL to download
    :param tag: The tag to extract the content from

    """
    tag_start = f'<{tag}>'
    tag_end = f'</{tag}>'
    is_header = False
    content = ""

    for response in stream_with_retry(url):
        for line in response.iter_lines():
            if line:
                # If line matches header_start, start capturing
                if line.startswith(tag_start):
                    is_header = True
                    continue  # Skip the current line as it's the opening tag

                # If line matches header_end, stop capturing
                elif line.startswith(tag_end):
                    break

                # If within header lines, add to header_content
                elif is_header:
                    content += line + '\n'  # Add a newline to preserve original line breaks
    return content


async def download_bulk_data(data_url: str) -> Path:
    """
    Download bulk data e.g. company facts, daily index, etc. from the SEC website
    e.g. "https://www.sec.gov/Archives/edgar/daily-index/xbrl/companyfacts.zip"
    """
    filename = os.path.basename(data_url)
    local_dir = filename.split('.')[0]
    download_path = edgar_data_dir / local_dir
    download_filename = download_path / filename

    # Create the directory if it doesn't exist
    if not download_path.exists():
        download_path.mkdir()

    if filename.endswith(".zip"):
        # Now stream the file to the data directory
        await stream_file(data_url, path=download_path)
        # Unzip the file to the data directory / file
        with zipfile.ZipFile(download_filename, 'r') as z:
            z.extractall(download_path)
        # Delete the zip file
        download_filename.unlink()
    return download_path


def download_datafile(data_url: str, local_directory:Path=None) -> Path:
    """Download a file to the local storage directory"""
    filename = os.path.basename(data_url)
    # Create the directory if it doesn't exist
    local_directory = local_directory or edgar_data_dir
    if not local_directory.exists():
        local_directory.mkdir()

    download_filename = local_directory / filename
    download_file(data_url, path=download_filename)
    return download_filename